"""Contains the Frame class."""

import _zivid
from zivid.camera_info import _to_camera_info
from zivid.camera_state import _to_camera_state
from zivid.frame_info import _to_frame_info
from zivid.image import Image
from zivid.settings2d import _to_settings2d


class Frame2D:
    """A 2D frame captured by a Zivid camera.

    Contains a 2D image as well as metadata, settings and state of the API at the time of capture.
    """

    def __init__(self, impl):
        """Initialize Frame2D wrapper.

        This constructor is only used internally, and should not be called by the end-user.

        Args:
            impl:   Reference to internal/back-end instance.

        Raises:
            TypeError: If argument does not match the expected internal class.
        """
        if isinstance(impl, _zivid.Frame2D):
            self.__impl = impl
        else:
            raise TypeError(
                "Unsupported type for argument impl. Got {}, expected {}.".format(type(impl), _zivid.Frame2D)
            )

    def __str__(self):
        return str(self.__impl)

    def image_rgba(self):
        """Get color (RGBA) image from the frame.

        Returns:
            An image instance containing RGBA data
        """
        return Image(self.__impl.image_rgba())

    def image_bgra(self):
        """Get color (BGRA) image from the frame.

        Returns:
            An image instance containing BGRA data
        """
        return Image(self.__impl.image_bgra())

    def image_rgba_srgb(self):
        """Get color (RGBA) image from the frame in the sRGB color space.

        Returns:
            An image instance containing RGBA data in sRGB color space
        """
        return Image(self.__impl.image_rgba_srgb())

    def image_bgra_srgb(self):
        """Get color (BGRA) image from the frame in the sRGB color space.

        Returns:
            An image instance containing BGRA data in sRGB color space
        """
        return Image(self.__impl.image_bgra_srgb())

    def image_srgb(self):
        """Get color (RGBA) image from the frame in the sRGB color space.

        This method is deprecated. Use image_rgba_srgb() instead.

        Returns:
            An image instance containing RGBA data in sRGB color space
        """
        return Image(self.__impl.image_rgba_srgb())

    @property
    def settings(self):
        """Get the settings used to capture this frame.

        Returns:
            A Settings2D instance
        """
        return _to_settings2d(self.__impl.settings)

    @property
    def state(self):
        """Get the camera state data at the time of the frame capture.

        Returns:
            A CameraState instance
        """
        return _to_camera_state(self.__impl.state)

    @property
    def info(self):
        """Get information collected at the time of the capture.

        Returns:
            A FrameInfo instance
        """
        return _to_frame_info(self.__impl.info)

    @property
    def camera_info(self):
        """Get information about the camera used to capture the frame.

        Returns:
            A CameraInfo instance
        """
        return _to_camera_info(self.__impl.camera_info)

    def release(self):
        """Release the underlying resources."""
        try:
            impl = self.__impl
        except AttributeError:
            pass
        else:
            impl.release()

    def clone(self):
        """Get a clone of the frame.

        The clone will include a copy of all the frame data.

        This function incurs a performance cost due to the copying of the data. When performance is important we
        recommend to avoid using this method, and instead modify the existing frame.

        This method is equivalent to calling `copy.deepcopy` on the frame. You can obtain a shallow copy that does
        not copy the underlying data by using `copy.copy` on the frame instead.

        Returns:
            A Frame2D instance
        """
        return Frame2D(self.__impl.clone())

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.release()

    def __del__(self):
        self.release()

    def __copy__(self):
        return Frame2D(self.__impl.__copy__())

    def __deepcopy__(self, memodict):
        return Frame2D(self.__impl.__deepcopy__(memodict))



import copy

import numpy as np
import pytest
import zivid
from zivid.camera_info import CameraInfo


def test_image_context_manager(frame_2d):
    with frame_2d.image_rgba() as image_rgba:
        assert image_rgba is not None
        assert isinstance(image_rgba, zivid.Image)

    with frame_2d.image_bgra() as image_bgra:
        assert image_bgra is not None
        assert isinstance(image_bgra, zivid.Image)


def test_image(frame_2d):
    image_rgba = frame_2d.image_rgba()
    assert image_rgba is not None
    assert isinstance(image_rgba, zivid.Image)

    image_bgra = frame_2d.image_bgra()
    assert image_bgra is not None
    assert isinstance(image_bgra, zivid.Image)

    image_rgba_srgb = frame_2d.image_rgba_srgb()
    assert image_rgba_srgb is not None
    assert isinstance(image_rgba_srgb, zivid.Image)

    image_bgra_srgb = frame_2d.image_bgra_srgb()
    assert image_bgra_srgb is not None
    assert isinstance(image_bgra_srgb, zivid.Image)

    image_srgb = frame_2d.image_srgb()
    assert image_srgb is not None
    assert isinstance(image_srgb, zivid.Image)


def test_deprecated_srgb(frame_2d):
    image_rgba_srgb = frame_2d.image_rgba_srgb()
    image_srgb = frame_2d.image_srgb()
    np.testing.assert_array_equal(image_rgba_srgb.copy_data(), image_srgb.copy_data())


def test_image_rgba_bgra_correspondence(frame_2d):
    rgba_linear = frame_2d.image_rgba().copy_data()
    bgra_linear = frame_2d.image_bgra().copy_data()

    rgba_srgb = frame_2d.image_rgba_srgb().copy_data()
    bgra_srgb = frame_2d.image_bgra_srgb().copy_data()

    for rgba, bgra in [(rgba_linear, bgra_linear), (rgba_srgb, bgra_srgb)]:
        np.testing.assert_array_equal(bgra[:, :, 0], rgba[:, :, 2])
        np.testing.assert_array_equal(bgra[:, :, 1], rgba[:, :, 1])
        np.testing.assert_array_equal(bgra[:, :, 2], rgba[:, :, 0])
        np.testing.assert_array_equal(bgra[:, :, 3], rgba[:, :, 3])


def test_state(frame_2d):
    state = frame_2d.state
    assert state is not None
    assert isinstance(state, zivid.CameraState)


def test_info(frame_2d):
    info = frame_2d.info
    assert info is not None
    assert isinstance(info, zivid.FrameInfo)


def test_camera_info(frame_2d):
    camera_info = frame_2d.camera_info
    assert camera_info
    assert isinstance(camera_info, CameraInfo)


def test_settings(frame_2d):
    settings_2d = frame_2d.settings
    assert settings_2d is not None
    assert isinstance(settings_2d, zivid.Settings2D)


def test_release(frame_2d):
    frame_2d.image_rgba()
    frame_2d.release()
    with pytest.raises(RuntimeError):
        frame_2d.image_rgba()


def test_context_manager(shared_file_camera):
    settings_2d = zivid.Settings2D(acquisitions=[zivid.Settings2D.Acquisition()])
    with shared_file_camera.capture(settings_2d) as frame_2d:
        frame_2d.image_rgba()
    with pytest.raises(RuntimeError):
        frame_2d.image_rgba()

    with shared_file_camera.capture(settings_2d) as frame_2d:
        frame_2d.image_bgra()
    with pytest.raises(RuntimeError):
        frame_2d.image_bgra()


def test_copy(frame_2d):
    with copy.copy(frame_2d) as frame_2d_copy:
        assert frame_2d_copy
        assert frame_2d_copy is not frame_2d
        assert isinstance(frame_2d_copy, zivid.Frame2D)
        np.testing.assert_array_equal(frame_2d.image_rgba().copy_data(), frame_2d_copy.image_rgba().copy_data())


def test_deepcopy(frame_2d):
    with copy.deepcopy(frame_2d) as frame_2d_copy:
        assert frame_2d_copy
        assert frame_2d_copy is not frame_2d
        assert isinstance(frame_2d_copy, zivid.Frame2D)
        np.testing.assert_array_equal(frame_2d.image_rgba().copy_data(), frame_2d_copy.image_rgba().copy_data())


def test_clone(frame_2d):
    frame_2d_clone = frame_2d.clone()
    assert frame_2d_clone
    assert frame_2d_clone is not frame_2d
    assert isinstance(frame_2d_clone, zivid.Frame2D)
    np.testing.assert_array_equal(frame_2d.image_rgba().copy_data(), frame_2d_clone.image_rgba().copy_data())



"""Auto generated, do not edit."""

# pylint: disable=too-many-lines,protected-access,too-few-public-methods,too-many-arguments,too-many-positional-arguments,line-too-long,missing-function-docstring,missing-class-docstring,redefined-builtin,too-many-branches,too-many-boolean-expressions
import _zivid


class LocalPointCloudRegistrationParameters:

    class ConvergenceCriteria:

        def __init__(
            self,
            rmse_diff_threshold=_zivid.toolbox.LocalPointCloudRegistrationParameters.ConvergenceCriteria.RMSEDiffThreshold().value,
            source_coverage_diff_threshold=_zivid.toolbox.LocalPointCloudRegistrationParameters.ConvergenceCriteria.SourceCoverageDiffThreshold().value,
        ):

            if isinstance(
                rmse_diff_threshold,
                (
                    float,
                    int,
                ),
            ):
                self._rmse_diff_threshold = (
                    _zivid.toolbox.LocalPointCloudRegistrationParameters.ConvergenceCriteria.RMSEDiffThreshold(
                        rmse_diff_threshold
                    )
                )
            else:
                raise TypeError(
                    "Unsupported type, expected: (float, int,), got {value_type}".format(
                        value_type=type(rmse_diff_threshold)
                    )
                )

            if isinstance(
                source_coverage_diff_threshold,
                (
                    float,
                    int,
                ),
            ):
                self._source_coverage_diff_threshold = _zivid.toolbox.LocalPointCloudRegistrationParameters.ConvergenceCriteria.SourceCoverageDiffThreshold(
                    source_coverage_diff_threshold
                )
            else:
                raise TypeError(
                    "Unsupported type, expected: (float, int,), got {value_type}".format(
                        value_type=type(source_coverage_diff_threshold)
                    )
                )

        @property
        def rmse_diff_threshold(self):
            return self._rmse_diff_threshold.value

        @property
        def source_coverage_diff_threshold(self):
            return self._source_coverage_diff_threshold.value

        @rmse_diff_threshold.setter
        def rmse_diff_threshold(self, value):
            if isinstance(
                value,
                (
                    float,
                    int,
                ),
            ):
                self._rmse_diff_threshold = (
                    _zivid.toolbox.LocalPointCloudRegistrationParameters.ConvergenceCriteria.RMSEDiffThreshold(value)
                )
            else:
                raise TypeError(
                    "Unsupported type, expected: float or  int, got {value_type}".format(value_type=type(value))
                )

        @source_coverage_diff_threshold.setter
        def source_coverage_diff_threshold(self, value):
            if isinstance(
                value,
                (
                    float,
                    int,
                ),
            ):
                self._source_coverage_diff_threshold = _zivid.toolbox.LocalPointCloudRegistrationParameters.ConvergenceCriteria.SourceCoverageDiffThreshold(
                    value
                )
            else:
                raise TypeError(
                    "Unsupported type, expected: float or  int, got {value_type}".format(value_type=type(value))
                )

        def __eq__(self, other):
            if (
                self._rmse_diff_threshold == other._rmse_diff_threshold
                and self._source_coverage_diff_threshold == other._source_coverage_diff_threshold
            ):
                return True
            return False

        def __str__(self):
            return str(_to_internal_toolbox_local_point_cloud_registration_parameters_convergence_criteria(self))

    def __init__(
        self,
        max_correspondence_distance=_zivid.toolbox.LocalPointCloudRegistrationParameters.MaxCorrespondenceDistance().value,
        max_iteration_count=_zivid.toolbox.LocalPointCloudRegistrationParameters.MaxIterationCount().value,
        convergence_criteria=None,
    ):

        if isinstance(
            max_correspondence_distance,
            (
                float,
                int,
            ),
        ):
            self._max_correspondence_distance = (
                _zivid.toolbox.LocalPointCloudRegistrationParameters.MaxCorrespondenceDistance(
                    max_correspondence_distance
                )
            )
        else:
            raise TypeError(
                "Unsupported type, expected: (float, int,), got {value_type}".format(
                    value_type=type(max_correspondence_distance)
                )
            )

        if isinstance(max_iteration_count, (int,)):
            self._max_iteration_count = _zivid.toolbox.LocalPointCloudRegistrationParameters.MaxIterationCount(
                max_iteration_count
            )
        else:
            raise TypeError(
                "Unsupported type, expected: (int,), got {value_type}".format(value_type=type(max_iteration_count))
            )

        if convergence_criteria is None:
            convergence_criteria = self.ConvergenceCriteria()
        if not isinstance(convergence_criteria, self.ConvergenceCriteria):
            raise TypeError("Unsupported type: {value}".format(value=type(convergence_criteria)))
        self._convergence_criteria = convergence_criteria

    @property
    def max_correspondence_distance(self):
        return self._max_correspondence_distance.value

    @property
    def max_iteration_count(self):
        return self._max_iteration_count.value

    @property
    def convergence_criteria(self):
        return self._convergence_criteria

    @max_correspondence_distance.setter
    def max_correspondence_distance(self, value):
        if isinstance(
            value,
            (
                float,
                int,
            ),
        ):
            self._max_correspondence_distance = (
                _zivid.toolbox.LocalPointCloudRegistrationParameters.MaxCorrespondenceDistance(value)
            )
        else:
            raise TypeError(
                "Unsupported type, expected: float or  int, got {value_type}".format(value_type=type(value))
            )

    @max_iteration_count.setter
    def max_iteration_count(self, value):
        if isinstance(value, (int,)):
            self._max_iteration_count = _zivid.toolbox.LocalPointCloudRegistrationParameters.MaxIterationCount(value)
        else:
            raise TypeError("Unsupported type, expected: int, got {value_type}".format(value_type=type(value)))

    @convergence_criteria.setter
    def convergence_criteria(self, value):
        if not isinstance(value, self.ConvergenceCriteria):
            raise TypeError("Unsupported type {value}".format(value=type(value)))
        self._convergence_criteria = value

    @classmethod
    def load(cls, file_name):
        return _to_toolbox_local_point_cloud_registration_parameters(
            _zivid.toolbox.LocalPointCloudRegistrationParameters(str(file_name))
        )

    def save(self, file_name):
        _to_internal_toolbox_local_point_cloud_registration_parameters(self).save(str(file_name))

    @classmethod
    def from_serialized(cls, value):
        return _to_toolbox_local_point_cloud_registration_parameters(
            _zivid.toolbox.LocalPointCloudRegistrationParameters.from_serialized(str(value))
        )

    def serialize(self):
        return _to_internal_toolbox_local_point_cloud_registration_parameters(self).serialize()

    def __eq__(self, other):
        if (
            self._max_correspondence_distance == other._max_correspondence_distance
            and self._max_iteration_count == other._max_iteration_count
            and self._convergence_criteria == other._convergence_criteria
        ):
            return True
        return False

    def __str__(self):
        return str(_to_internal_toolbox_local_point_cloud_registration_parameters(self))


def _to_toolbox_local_point_cloud_registration_parameters_convergence_criteria(internal_convergence_criteria):
    return LocalPointCloudRegistrationParameters.ConvergenceCriteria(
        rmse_diff_threshold=internal_convergence_criteria.rmse_diff_threshold.value,
        source_coverage_diff_threshold=internal_convergence_criteria.source_coverage_diff_threshold.value,
    )


def _to_toolbox_local_point_cloud_registration_parameters(internal_local_point_cloud_registration_parameters):
    return LocalPointCloudRegistrationParameters(
        convergence_criteria=_to_toolbox_local_point_cloud_registration_parameters_convergence_criteria(
            internal_local_point_cloud_registration_parameters.convergence_criteria
        ),
        max_correspondence_distance=internal_local_point_cloud_registration_parameters.max_correspondence_distance.value,
        max_iteration_count=internal_local_point_cloud_registration_parameters.max_iteration_count.value,
    )


def _to_internal_toolbox_local_point_cloud_registration_parameters_convergence_criteria(convergence_criteria):
    internal_convergence_criteria = _zivid.toolbox.LocalPointCloudRegistrationParameters.ConvergenceCriteria()

    internal_convergence_criteria.rmse_diff_threshold = (
        _zivid.toolbox.LocalPointCloudRegistrationParameters.ConvergenceCriteria.RMSEDiffThreshold(
            convergence_criteria.rmse_diff_threshold
        )
    )
    internal_convergence_criteria.source_coverage_diff_threshold = (
        _zivid.toolbox.LocalPointCloudRegistrationParameters.ConvergenceCriteria.SourceCoverageDiffThreshold(
            convergence_criteria.source_coverage_diff_threshold
        )
    )

    return internal_convergence_criteria


def _to_internal_toolbox_local_point_cloud_registration_parameters(local_point_cloud_registration_parameters):
    internal_local_point_cloud_registration_parameters = _zivid.toolbox.LocalPointCloudRegistrationParameters()

    internal_local_point_cloud_registration_parameters.max_correspondence_distance = (
        _zivid.toolbox.LocalPointCloudRegistrationParameters.MaxCorrespondenceDistance(
            local_point_cloud_registration_parameters.max_correspondence_distance
        )
    )
    internal_local_point_cloud_registration_parameters.max_iteration_count = (
        _zivid.toolbox.LocalPointCloudRegistrationParameters.MaxIterationCount(
            local_point_cloud_registration_parameters.max_iteration_count
        )
    )

    internal_local_point_cloud_registration_parameters.convergence_criteria = (
        _to_internal_toolbox_local_point_cloud_registration_parameters_convergence_criteria(
            local_point_cloud_registration_parameters.convergence_criteria
        )
    )
    return internal_local_point_cloud_registration_parameters



import copy
import tempfile
from pathlib import Path

import pytest
import zivid
from assertions import assert_point_clouds_equal, assert_point_clouds_not_equal
from zivid import CameraInfo, CameraState, FrameInfo, Settings


def test_illegal_init(
    application,  # pylint: disable=unused-argument
):
    with pytest.raises(RuntimeError):
        zivid.frame.Frame("non-exisiting-file.zdf")
    with pytest.raises(TypeError):
        zivid.frame.Frame(None)
    with pytest.raises(TypeError):
        zivid.frame.Frame(12345)


def test_point_cloud(frame):
    point_cloud = frame.point_cloud()
    assert isinstance(point_cloud, zivid.PointCloud)


def test_frame_2d(frame):
    frame_2d = frame.frame_2d()
    assert frame_2d
    assert isinstance(frame_2d, zivid.Frame2D)


def test_path_init(
    application,  # pylint: disable=unused-argument
    frame_file,
):
    frame = zivid.frame.Frame(frame_file)
    assert isinstance(frame_file, Path)
    assert frame is not None
    assert isinstance(frame, zivid.frame.Frame)


def test_str_as_path_init(
    application,  # pylint: disable=unused-argument
    frame_file,
):
    frame = zivid.frame.Frame(str(frame_file))
    assert frame is not None
    assert isinstance(frame, zivid.frame.Frame)


def test_save(frame):
    with tempfile.TemporaryDirectory() as temp_dir:
        save_path = Path(temp_dir) / "save_test.zdf"
        frame.save(save_path)
        assert save_path.exists()


def test_context_manager(
    application,  # pylint: disable=unused-argument
    frame_file,
):
    with zivid.frame.Frame(frame_file) as frame:
        frame.point_cloud()
    with pytest.raises(RuntimeError):
        frame.point_cloud()


def test_to_string(frame):
    string = str(frame)
    assert string
    assert isinstance(string, str)


def test_load(frame, frame_file):
    assert frame.load(frame_file) is None


def test_settings(frame):
    settings = frame.settings
    assert settings
    assert isinstance(settings, Settings)


def test_state(frame):
    state = frame.state
    assert state
    assert isinstance(state, CameraState)


def test_info(frame):
    info = frame.info
    assert info
    assert isinstance(info, FrameInfo)


def test_camera_info(frame):
    camera_info = frame.camera_info
    assert camera_info
    assert isinstance(camera_info, CameraInfo)


def test_release(frame):
    frame.point_cloud()
    frame.release()
    with pytest.raises(RuntimeError):
        frame.point_cloud()


def test_copy(frame, transform):
    frame_copy = copy.copy(frame)
    assert frame_copy
    assert frame_copy is not frame
    assert isinstance(frame_copy, type(frame))

    assert_point_clouds_equal(frame.point_cloud(), frame_copy.point_cloud())
    # shallow copy, point cloud transform should affect both
    frame.point_cloud().transform(transform)
    assert_point_clouds_equal(frame.point_cloud(), frame_copy.point_cloud())

    frame.release()
    assert isinstance(frame_copy.point_cloud(), zivid.PointCloud)


def test_deepcopy(frame, transform):
    frame_deepcopy = copy.deepcopy(frame)
    assert frame_deepcopy
    assert frame_deepcopy is not frame
    assert isinstance(frame_deepcopy, type(frame))

    assert_point_clouds_equal(frame.point_cloud(), frame_deepcopy.point_cloud())
    # deep copy, point cloud transform should not affect both
    frame.point_cloud().transform(transform)
    assert_point_clouds_not_equal(frame.point_cloud(), frame_deepcopy.point_cloud())

    frame.release()
    assert isinstance(frame_deepcopy.point_cloud(), zivid.PointCloud)


def test_clone(frame, transform):
    frame_clone = frame.clone()
    assert frame_clone
    assert frame_clone is not frame
    assert isinstance(frame_clone, type(frame))

    assert_point_clouds_equal(frame.point_cloud(), frame_clone.point_cloud())
    # clone, point cloud transform should not affect both
    frame.point_cloud().transform(transform)
    assert_point_clouds_not_equal(frame.point_cloud(), frame_clone.point_cloud())

    frame.release()
    assert isinstance(frame_clone.point_cloud(), zivid.PointCloud)



import datetime
import numbers

import pytest
import zivid
from zivid import Settings, Settings2D
from zivid.settings import _to_internal_settings, _to_settings


def test_default_settings(
    application,  # pylint: disable=unused-argument
):
    settings = zivid.Settings()

    assert settings.color is None

    assert isinstance(settings.acquisitions, list)
    assert len(settings.acquisitions) == 0
    assert settings.engine is None
    assert isinstance(settings.diagnostics, zivid.Settings.Diagnostics)
    assert settings.diagnostics.enabled is None
    assert isinstance(settings.processing, zivid.Settings.Processing)
    assert isinstance(settings.processing.color, zivid.Settings.Processing.Color)
    assert isinstance(settings.processing.color.balance, zivid.Settings.Processing.Color.Balance)
    assert settings.processing.color.gamma is None
    assert settings.processing.color.balance.red is None
    assert settings.processing.color.balance.green is None
    assert settings.processing.color.balance.blue is None
    assert settings.processing.color.experimental.mode is None

    assert isinstance(settings.processing.filters, zivid.Settings.Processing.Filters)
    assert isinstance(
        settings.processing.filters.experimental,
        zivid.Settings.Processing.Filters.Experimental,
    )

    assert isinstance(settings.processing.filters.cluster, zivid.Settings.Processing.Filters.Cluster)
    assert isinstance(
        settings.processing.filters.cluster.removal,
        zivid.Settings.Processing.Filters.Cluster.Removal,
    )

    assert settings.processing.filters.cluster.removal.enabled is None
    assert settings.processing.filters.cluster.removal.min_area is None
    assert settings.processing.filters.cluster.removal.max_neighbor_distance is None

    assert isinstance(
        settings.processing.filters.experimental.contrast_distortion,
        zivid.Settings.Processing.Filters.Experimental.ContrastDistortion,
    )
    assert isinstance(
        settings.processing.filters.experimental.contrast_distortion.correction,
        zivid.Settings.Processing.Filters.Experimental.ContrastDistortion.Correction,
    )
    assert settings.processing.filters.experimental.contrast_distortion.correction.enabled is None

    assert isinstance(
        settings.processing.filters.experimental.contrast_distortion.removal,
        zivid.Settings.Processing.Filters.Experimental.ContrastDistortion.Removal,
    )
    assert settings.processing.filters.experimental.contrast_distortion.removal.enabled is None

    assert isinstance(
        settings.processing.filters.hole.repair,
        zivid.Settings.Processing.Filters.Hole.Repair,
    )

    assert settings.processing.filters.hole.repair.enabled is None
    assert settings.processing.filters.hole.repair.hole_size is None
    assert settings.processing.filters.hole.repair.strictness is None

    assert isinstance(settings.processing.filters.noise, zivid.Settings.Processing.Filters.Noise)
    assert isinstance(
        settings.processing.filters.noise.removal,
        zivid.Settings.Processing.Filters.Noise.Removal,
    )
    assert settings.processing.filters.noise.removal.enabled is None

    assert isinstance(settings.processing.filters.outlier, zivid.Settings.Processing.Filters.Outlier)
    assert isinstance(
        settings.processing.filters.outlier.removal,
        zivid.Settings.Processing.Filters.Outlier.Removal,
    )
    assert settings.processing.filters.outlier.removal.enabled is None

    assert isinstance(
        settings.processing.filters.reflection,
        zivid.Settings.Processing.Filters.Reflection,
    )
    assert isinstance(
        settings.processing.filters.reflection.removal,
        zivid.Settings.Processing.Filters.Reflection.Removal,
    )
    assert settings.processing.filters.reflection.removal.enabled is None
    assert settings.processing.filters.reflection.removal.mode is None

    assert isinstance(
        settings.processing.filters.smoothing,
        zivid.Settings.Processing.Filters.Smoothing,
    )
    assert isinstance(
        settings.processing.filters.smoothing.gaussian,
        zivid.Settings.Processing.Filters.Smoothing.Gaussian,
    )
    assert settings.processing.filters.smoothing.gaussian.enabled is None
    assert settings.processing.filters.smoothing.gaussian.sigma is None

    assert isinstance(settings.region_of_interest, zivid.Settings.RegionOfInterest)
    assert isinstance(settings.region_of_interest.box, zivid.Settings.RegionOfInterest.Box)
    assert isinstance(settings.region_of_interest.depth, zivid.Settings.RegionOfInterest.Depth)

    assert settings.region_of_interest.box.enabled is None
    assert settings.region_of_interest.box.extents is None
    assert settings.region_of_interest.box.point_o is None
    assert settings.region_of_interest.box.point_a is None
    assert settings.region_of_interest.box.point_b is None
    assert settings.region_of_interest.depth.enabled is None
    assert settings.region_of_interest.depth.range is None


def test_set_color_settings():
    # pylint: disable=protected-access

    settings = Settings()
    assert settings.color is None

    to_cpp_and_back = _to_settings(_to_internal_settings(settings))
    assert to_cpp_and_back.color is None
    assert to_cpp_and_back == settings

    settings.color = Settings2D()
    assert settings.color is not None
    assert isinstance(settings.color, Settings2D)
    assert settings.color == Settings2D()
    to_cpp_and_back = _to_settings(_to_internal_settings(settings))
    assert to_cpp_and_back.color is not None
    assert to_cpp_and_back == settings

    settings = Settings(color=Settings2D())
    assert settings.color is not None
    assert isinstance(settings.color, Settings2D)
    assert settings.color == Settings2D()
    to_cpp_and_back = _to_settings(_to_internal_settings(settings))
    assert to_cpp_and_back.color is not None
    assert to_cpp_and_back == settings

    settings = Settings(color=Settings2D(acquisitions=(Settings2D.Acquisition(),)))
    assert settings.color is not None
    assert isinstance(settings.color, Settings2D)
    assert len(settings.color.acquisitions) == 1
    to_cpp_and_back = _to_settings(_to_internal_settings(settings))
    assert to_cpp_and_back.color is not None
    assert to_cpp_and_back == settings

    settings = Settings()
    settings.color = Settings2D()
    settings.color.acquisitions.append(Settings2D.Acquisition())
    assert len(settings.color.acquisitions) == 1
    to_cpp_and_back = _to_settings(_to_internal_settings(settings))
    assert len(to_cpp_and_back.color.acquisitions) == 1
    assert to_cpp_and_back == settings


def test_set_acquisition_list():
    settings = Settings()

    settings.acquisitions = [
        Settings.Acquisition(gain=1.0),
        Settings.Acquisition(gain=2.0),
        Settings.Acquisition(gain=3.0),
    ]
    assert len(settings.acquisitions) == 3
    assert settings.acquisitions is not None
    assert isinstance(settings.acquisitions, list)
    for element in settings.acquisitions:
        assert isinstance(element, Settings.Acquisition)

    assert settings.acquisitions[0].gain == 1.0
    assert settings.acquisitions[1].gain == 2.0
    assert settings.acquisitions[2].gain == 3.0

    settings.acquisitions[0].gain = 4.0
    assert settings.acquisitions[0].gain == 4.0


def test_set_acquisition_generator():
    settings = Settings()

    def _generator():
        for _ in range(3):
            yield Settings.Acquisition()

    settings.acquisitions = _generator()
    assert settings.acquisitions is not None
    assert isinstance(settings.acquisitions, list)
    for element in settings.acquisitions:
        assert isinstance(element, Settings.Acquisition)


def test_set_acquisition_tuple():
    settings = Settings()

    settings.acquisitions = (Settings.Acquisition(), settings.Acquisition())
    assert settings.acquisitions is not None
    assert isinstance(settings.acquisitions, list)
    for element in settings.acquisitions:
        assert isinstance(element, Settings.Acquisition)


def test_default_acquisition(
    application,  # pylint: disable=unused-argument
):
    settings = zivid.Settings(acquisitions=[zivid.Settings.Acquisition()])
    assert isinstance(settings.acquisitions, list)
    acquisition = settings.acquisitions[0]

    assert isinstance(acquisition, zivid.Settings.Acquisition)
    assert acquisition.aperture is None
    assert acquisition.brightness is None
    assert acquisition.gain is None
    assert acquisition.exposure_time is None
    pytest.helpers.equality_tester(
        zivid.Settings.Acquisition,
        [5, 0.5, datetime.timedelta(microseconds=11000), 15],
        [5, 0.5, datetime.timedelta(microseconds=11001), 15],
    )


def test_acquisition_brightness(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.Acquisition(),
        member="brightness",
        value=0.5,
        expected_data_type=numbers.Real,
    )


def test_acquisition_exposure_time(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.Acquisition(),
        member="exposure_time",
        value=datetime.timedelta(microseconds=100000),
        expected_data_type=datetime.timedelta,
    )


def test_acquisition_gain(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.Acquisition(),
        member="gain",
        value=14,
        expected_data_type=numbers.Real,
    )


def test_acquisition_aperture(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.Acquisition(),
        member="aperture",
        value=20.5,
        expected_data_type=numbers.Real,
    )


def test_settings_diagnostics(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings(),
        member="diagnostics",
        value=zivid.Settings.Diagnostics(),
        expected_data_type=zivid.Settings.Diagnostics,
    )
    pytest.helpers.equality_tester(
        zivid.Settings.Diagnostics,
        [True],
        [False],
    )


def test_settings_diagnostics_enabled(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.Diagnostics(),
        member="enabled",
        value=True,
        expected_data_type=bool,
    )


def test_settings_engine(
    application,  # pylint: disable=unused-argument
):
    for value in zivid.Settings.Engine.valid_values():
        pytest.helpers.set_attribute_tester(
            settings_instance=zivid.Settings(),
            member="engine",
            value=value,
            expected_data_type=str,
        )
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings(),
        member="engine",
        value=None,
        expected_data_type=type(None),
    )
    # Is optional enum
    zivid.Settings(engine="stripe")
    zivid.Settings(engine="phase")
    zivid.Settings(engine=None)
    with pytest.raises(KeyError):
        zivid.Settings(engine="_dummy_")


def test_settings_processing(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings(),
        member="processing",
        value=zivid.Settings.Processing(),
        expected_data_type=zivid.Settings.Processing,
    )
    pytest.helpers.equality_tester(
        zivid.Settings.Processing,
        [zivid.Settings.Processing.Color(0.9, zivid.Settings.Processing.Color.Balance(blue=1.1))],
        [zivid.Settings.Processing.Color(1.1, zivid.Settings.Processing.Color.Balance(blue=1.2))],
    )


def test_settings_processing_color(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.Processing(),
        member="color",
        value=zivid.Settings.Processing.Color(),
        expected_data_type=zivid.Settings.Processing.Color,
    )
    pytest.helpers.equality_tester(
        zivid.Settings.Processing.Color,
        [0.9, zivid.Settings.Processing.Color.Balance(blue=1.1)],
        [1.1, zivid.Settings.Processing.Color.Balance(blue=1.2)],
    )


def test_settings_processing_color_gamma(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.Processing.Color(),
        member="gamma",
        value=0.85,
        expected_data_type=numbers.Real,
    )


def test_settings_processing_color_balance(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.Processing.Color(),
        member="balance",
        value=zivid.Settings.Processing.Color.Balance(),
        expected_data_type=zivid.Settings.Processing.Color.Balance,
    )
    pytest.helpers.equality_tester(
        zivid.Settings.Processing.Color.Balance,
        [1.1, 1.1, 1.1],
        [1.2, 1.1, 1.1],
    )


def test_settings_processing_color_experimental(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.Processing.Color(),
        member="experimental",
        value=zivid.Settings.Processing.Color.Experimental(),
        expected_data_type=zivid.Settings.Processing.Color.Experimental,
    )

    pytest.helpers.equality_tester(
        zivid.Settings.Processing.Color.Experimental,
        ["automatic"],
        [None],
    )

    pytest.helpers.equality_tester(
        zivid.Settings.Processing.Color.Experimental,
        ["automatic"],
        ["useFirstAcquisition"],
    )

    pytest.helpers.equality_tester(
        zivid.Settings.Processing.Color.Experimental,
        ["automatic"],
        ["toneMapping"],
    )


def test_settings_processing_color_experimental_mode(
    application,  # pylint: disable=unused-argument
):
    for value in zivid.Settings.Processing.Color.Experimental.Mode.valid_values():
        pytest.helpers.set_attribute_tester(
            settings_instance=zivid.Settings.Processing.Color.Experimental(),
            member="mode",
            value=value,
            expected_data_type=str,
        )
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.Processing.Color.Experimental(),
        member="mode",
        value=None,
        expected_data_type=type(None),
    )
    # Is optional enum
    zivid.Settings.Processing.Color.Experimental(mode="automatic")
    zivid.Settings.Processing.Color.Experimental(mode="useFirstAcquisition")
    zivid.Settings.Processing.Color.Experimental(mode="toneMapping")
    zivid.Settings.Processing.Color.Experimental(mode=None)
    with pytest.raises(KeyError):
        zivid.Settings.Processing.Color.Experimental(mode="_dummy_")


def test_settings_processing_color_balance_red(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.Processing.Color.Balance(),
        member="red",
        value=2,
        expected_data_type=numbers.Real,
    )


def test_settings_processing_color_balance_green(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.Processing.Color.Balance(),
        member="green",
        value=1.5,
        expected_data_type=numbers.Real,
    )


def test_settings_processing_color_balance_blue(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.Processing.Color.Balance(),
        member="blue",
        value=1,
        expected_data_type=numbers.Real,
    )


def test_settings_processing_filters(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.Processing(),
        member="filters",
        value=zivid.Settings.Processing.Filters(),
        expected_data_type=zivid.Settings.Processing.Filters,
    )
    pytest.helpers.equality_tester(
        zivid.Settings.Processing.Filters,
        [zivid.Settings.Processing.Filters.Cluster(zivid.Settings.Processing.Filters.Cluster.Removal(enabled=True))],
        [zivid.Settings.Processing.Filters.Cluster(zivid.Settings.Processing.Filters.Cluster.Removal(enabled=False))],
    )


def test_settings_processing_filters_experimental(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.Processing.Filters(),
        member="experimental",
        value=zivid.Settings.Processing.Filters.Experimental(),
        expected_data_type=zivid.Settings.Processing.Filters.Experimental,
    )
    pytest.helpers.equality_tester(
        zivid.Settings.Processing.Filters.Experimental,
        [
            zivid.Settings.Processing.Filters.Experimental.ContrastDistortion(
                removal=zivid.Settings.Processing.Filters.Experimental.ContrastDistortion.Removal(enabled=False)
            )
        ],
        [
            zivid.Settings.Processing.Filters.Experimental.ContrastDistortion(
                removal=zivid.Settings.Processing.Filters.Experimental.ContrastDistortion.Removal(enabled=True)
            )
        ],
    )


def test_settings_processing_filters_experimental_contrast_distortion(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.Processing.Filters.Experimental(),
        member="contrast_distortion",
        value=zivid.Settings.Processing.Filters.Experimental.ContrastDistortion(),
        expected_data_type=zivid.Settings.Processing.Filters.Experimental.ContrastDistortion,
    )
    pytest.helpers.equality_tester(
        zivid.Settings.Processing.Filters.Experimental.ContrastDistortion,
        [
            zivid.Settings.Processing.Filters.Experimental.ContrastDistortion.Correction(enabled=True),
            zivid.Settings.Processing.Filters.Experimental.ContrastDistortion.Removal(enabled=False),
        ],
        [
            zivid.Settings.Processing.Filters.Experimental.ContrastDistortion.Correction(enabled=True),
            zivid.Settings.Processing.Filters.Experimental.ContrastDistortion.Removal(enabled=True),
        ],
    )


def test_settings_processing_filters_hole_repair(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.Processing.Filters.Hole(),
        member="repair",
        value=zivid.Settings.Processing.Filters.Hole.Repair(),
        expected_data_type=zivid.Settings.Processing.Filters.Hole.Repair,
    )

    pytest.helpers.equality_tester(
        zivid.Settings.Processing.Filters.Hole.Repair,
        [True],
        [False],
    )


def test_settings_processing_filters_hole_repair_enabled(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.Processing.Filters.Hole.Repair(),
        member="enabled",
        value=True,
        expected_data_type=bool,
    )


def test_settings_processing_filters_hole_repair_holesize(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.Processing.Filters.Hole.Repair(),
        member="hole_size",
        value=0.3,
        expected_data_type=float,
    )


def test_settings_processing_filters_hole_repair_strictness(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.Processing.Filters.Hole.Repair(),
        member="strictness",
        value=3,
        expected_data_type=int,
    )


def test_settings_processing_filters_experimental_contrast_distortion_removal(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.Processing.Filters.Experimental.ContrastDistortion(),
        member="removal",
        value=zivid.Settings.Processing.Filters.Experimental.ContrastDistortion.Removal(),
        expected_data_type=zivid.Settings.Processing.Filters.Experimental.ContrastDistortion.Removal,
    )
    pytest.helpers.equality_tester(
        zivid.Settings.Processing.Filters.Experimental.ContrastDistortion.Removal,
        [True],
        [False],
    )


def test_settings_processing_filters_experimental_contrast_distortion_removal_enabled(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.Processing.Filters.Experimental.ContrastDistortion.Removal(),
        member="enabled",
        value=True,
        expected_data_type=bool,
    )


def test_settings_processing_filters_experimental_contrast_distortion_removal_threshold(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.Processing.Filters.Experimental.ContrastDistortion.Removal(),
        member="threshold",
        value=0.4,
        expected_data_type=float,
    )


def test_settings_processing_filters_experimental_contrast_distortion_correction(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.Processing.Filters.Experimental.ContrastDistortion(),
        member="correction",
        value=zivid.Settings.Processing.Filters.Experimental.ContrastDistortion.Correction(),
        expected_data_type=zivid.Settings.Processing.Filters.Experimental.ContrastDistortion.Correction,
    )
    pytest.helpers.equality_tester(
        zivid.Settings.Processing.Filters.Experimental.ContrastDistortion.Correction,
        [True],
        [False],
    )


def test_settings_processing_filters_experimental_contrast_distortion_correction_strength(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.Processing.Filters.Experimental.ContrastDistortion.Correction(),
        member="strength",
        value=0.59,
        expected_data_type=float,
    )


def test_settings_processing_filters_experimental_contrast_distortion_correction_enabled(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.Processing.Filters.Experimental.ContrastDistortion.Correction(),
        member="enabled",
        value=False,
        expected_data_type=bool,
    )


def test_settings_processing_filters_cluster(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.Processing.Filters(),
        member="cluster",
        value=zivid.Settings.Processing.Filters.Cluster(),
        expected_data_type=zivid.Settings.Processing.Filters.Cluster,
    )

    pytest.helpers.equality_tester(
        zivid.Settings.Processing.Filters.Cluster,
        [zivid.Settings.Processing.Filters.Cluster.Removal(enabled=True)],
        [zivid.Settings.Processing.Filters.Cluster.Removal(enabled=False)],
    )


def test_settings_processing_filters_cluster_removal(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.Processing.Filters.Cluster(),
        member="removal",
        value=zivid.Settings.Processing.Filters.Cluster.Removal(),
        expected_data_type=zivid.Settings.Processing.Filters.Cluster.Removal,
    )
    pytest.helpers.equality_tester(
        zivid.Settings.Processing.Filters.Cluster.Removal,
        [True],
        [False],
    )


def test_settings_processing_filters_cluster_removal_enabled(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.Processing.Filters.Cluster.Removal(),
        member="enabled",
        value=False,
        expected_data_type=bool,
    )


def test_settings_processing_filters_cluster_removal_minarea(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.Processing.Filters.Cluster.Removal(),
        member="min_area",
        value=150.0,
        expected_data_type=float,
    )


def test_settings_processing_filters_cluster_removal_maxneighbordistance(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.Processing.Filters.Cluster.Removal(),
        member="max_neighbor_distance",
        value=5.0,
        expected_data_type=float,
    )


def test_settings_processing_filters_noise(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.Processing.Filters(),
        member="noise",
        value=zivid.Settings.Processing.Filters.Noise(),
        expected_data_type=zivid.Settings.Processing.Filters.Noise,
    )
    pytest.helpers.equality_tester(
        zivid.Settings.Processing.Filters.Noise,
        [zivid.Settings.Processing.Filters.Noise.Removal(enabled=True)],
        [zivid.Settings.Processing.Filters.Noise.Removal(enabled=False)],
    )


def test_settings_processing_filters_noise_removal(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.Processing.Filters.Noise(),
        member="removal",
        value=zivid.Settings.Processing.Filters.Noise.Removal(),
        expected_data_type=zivid.Settings.Processing.Filters.Noise.Removal,
    )
    pytest.helpers.equality_tester(
        zivid.Settings.Processing.Filters.Noise.Removal,
        [True],
        [False],
    )


def test_settings_processing_filters_noise_removal_enabled(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.Processing.Filters.Noise.Removal(),
        member="enabled",
        value=False,
        expected_data_type=bool,
    )


def test_settings_processing_filters_noise_removal_threshold(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.Processing.Filters.Noise.Removal(),
        member="threshold",
        value=50.0,
        expected_data_type=float,
    )


def test_settings_processing_filters_outlier(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.Processing.Filters(),
        member="outlier",
        value=zivid.Settings.Processing.Filters.Outlier(),
        expected_data_type=zivid.Settings.Processing.Filters.Outlier,
    )
    pytest.helpers.equality_tester(
        zivid.Settings.Processing.Filters.Outlier,
        [zivid.Settings.Processing.Filters.Outlier.Removal(enabled=True)],
        [zivid.Settings.Processing.Filters.Outlier.Removal(enabled=False)],
    )


def test_settings_processing_filters_outlier_removal(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.Processing.Filters.Outlier(),
        member="removal",
        value=zivid.Settings.Processing.Filters.Outlier.Removal(),
        expected_data_type=zivid.Settings.Processing.Filters.Outlier.Removal,
    )
    pytest.helpers.equality_tester(
        zivid.Settings.Processing.Filters.Outlier.Removal,
        [True],
        [False],
    )


def test_settings_processing_filters_outlier_removal_enabled(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.Processing.Filters.Outlier.Removal(),
        member="enabled",
        value=False,
        expected_data_type=bool,
    )


def test_settings_processing_filters_outlier_removal_threshold(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.Processing.Filters.Outlier.Removal(),
        member="threshold",
        value=89,
        expected_data_type=float,
    )


def test_settings_processing_filters_reflection(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.Processing.Filters(),
        member="reflection",
        value=zivid.Settings.Processing.Filters.Reflection(),
        expected_data_type=zivid.Settings.Processing.Filters.Reflection,
    )
    pytest.helpers.equality_tester(
        zivid.Settings.Processing.Filters.Reflection,
        [zivid.Settings.Processing.Filters.Reflection.Removal(enabled=True)],
        [zivid.Settings.Processing.Filters.Reflection.Removal(enabled=False)],
    )


def test_settings_processing_filters_reflection_removal(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.Processing.Filters.Reflection(),
        member="removal",
        value=zivid.Settings.Processing.Filters.Reflection.Removal(),
        expected_data_type=zivid.Settings.Processing.Filters.Reflection.Removal,
    )

    pytest.helpers.equality_tester(
        zivid.Settings.Processing.Filters.Reflection.Removal,
        [True],
        [False],
    )

    pytest.helpers.equality_tester(
        zivid.Settings.Processing.Filters.Reflection.Removal,
        [
            True,
            zivid.Settings.Processing.Filters.Reflection.Removal.Mode.global_,
        ],
        [
            True,
            zivid.Settings.Processing.Filters.Reflection.Removal.Mode.local,
        ],
    )


def test_settings_processing_filters_reflection_removal_enabled(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.Processing.Filters.Reflection.Removal(),
        member="enabled",
        value=True,
        expected_data_type=bool,
    )


def test_settings_processing_filters_reflection_removal_mode(
    application,  # pylint: disable=unused-argument
):
    for value in zivid.Settings.Processing.Filters.Reflection.Removal.Mode.valid_values():
        pytest.helpers.set_attribute_tester(
            settings_instance=zivid.Settings.Processing.Filters.Reflection.Removal(),
            member="mode",
            value=value,
            expected_data_type=str,
        )
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.Processing.Filters.Reflection.Removal(),
        member="mode",
        value=None,
        expected_data_type=type(None),
    )

    # Is optional enum
    zivid.Settings.Processing.Filters.Reflection.Removal(mode="global")
    zivid.Settings.Processing.Filters.Reflection.Removal(mode="local")
    with pytest.raises(KeyError):
        zivid.Settings.Processing.Filters.Reflection.Removal(mode="_dummy_")


def test_settings_processing_filters_smoothing(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.Processing.Filters(),
        member="smoothing",
        value=zivid.Settings.Processing.Filters.Smoothing(),
        expected_data_type=zivid.Settings.Processing.Filters.Smoothing,
    )
    pytest.helpers.equality_tester(
        zivid.Settings.Processing.Filters.Smoothing,
        [zivid.Settings.Processing.Filters.Smoothing.Gaussian(enabled=True)],
        [zivid.Settings.Processing.Filters.Smoothing.Gaussian(enabled=False)],
    )


def test_settings_processing_filters_smoothing_gaussian(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.Processing.Filters.Smoothing(),
        member="gaussian",
        value=zivid.Settings.Processing.Filters.Smoothing.Gaussian(),
        expected_data_type=zivid.Settings.Processing.Filters.Smoothing.Gaussian,
    )
    pytest.helpers.equality_tester(
        zivid.Settings.Processing.Filters.Smoothing.Gaussian,
        [True],
        [False],
    )


def test_settings_processing_filters_smoothing_gaussian_sigma(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.Processing.Filters.Smoothing.Gaussian(),
        member="sigma",
        value=1.798888,
        expected_data_type=float,
    )


def test_settings_processing_filters_smoothing_gaussian_enabled(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.Processing.Filters.Smoothing.Gaussian(),
        member="enabled",
        value=True,
        expected_data_type=bool,
    )


def test_settings_regionofinterest(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings(),
        member="region_of_interest",
        value=zivid.Settings.RegionOfInterest(),
        expected_data_type=zivid.Settings.RegionOfInterest,
    )
    pytest.helpers.equality_tester(
        zivid.Settings.RegionOfInterest,
        [zivid.Settings.RegionOfInterest.Box(enabled=True)],
        [zivid.Settings.RegionOfInterest.Box(enabled=False)],
    )


def test_settings_regionofinterest_box(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.RegionOfInterest(),
        member="box",
        value=zivid.Settings.RegionOfInterest.Box(),
        expected_data_type=zivid.Settings.RegionOfInterest.Box,
    )
    pytest.helpers.equality_tester(
        zivid.Settings.RegionOfInterest.Box,
        [True],
        [False],
    )


def test_settings_regionofinterest_box_enabled(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.RegionOfInterest.Box(),
        member="enabled",
        value=True,
        expected_data_type=bool,
    )


def test_settings_regionofinterest_box_pointo(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.RegionOfInterest.Box(),
        member="point_o",
        value=[100.0, 200.0, 300.0],
        expected_data_type=list,
    )


def test_settings_regionofinterest_box_pointa(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.RegionOfInterest.Box(),
        member="point_a",
        value=[100.0, 200.0, 300.0],
        expected_data_type=list,
    )


def test_settings_regionofinterest_box_pointb(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.RegionOfInterest.Box(),
        member="point_b",
        value=[100.0, 200.0, 300.0],
        expected_data_type=list,
    )


def test_settings_regionofinterest_box_extents(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.RegionOfInterest.Box(),
        member="extents",
        value=[-100.0, 500.0],
        expected_data_type=list,
    )

    # A range must have two elements
    with pytest.raises(TypeError):
        zivid.Settings.RegionOfInterest.Box(extents=[])
    with pytest.raises(TypeError):
        zivid.Settings.RegionOfInterest.Box(extents=[100.0])
    with pytest.raises(TypeError):
        zivid.Settings.RegionOfInterest.Box(extents=[100.0, 200.0, 300.0])

    # A range must have min <= max
    with pytest.raises(TypeError):
        zivid.Settings.RegionOfInterest.Box(extents=[100.0, 50.0])


def test_settings_regionofinterest_depth(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.RegionOfInterest(),
        member="depth",
        value=zivid.Settings.RegionOfInterest.Depth(),
        expected_data_type=zivid.Settings.RegionOfInterest.Depth,
    )
    pytest.helpers.equality_tester(
        zivid.Settings.RegionOfInterest.Depth,
        [True],
        [False],
    )


def test_settings_regionofinterest_depth_enabled(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.RegionOfInterest.Depth(),
        member="enabled",
        value=True,
        expected_data_type=bool,
    )


def test_settings_regionofinterest_depth_range(
    application,  # pylint: disable=unused-argument
):
    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings.RegionOfInterest.Depth(),
        member="range",
        value=[100.0, 1500.0],
        expected_data_type=list,
    )

    # A range must have two elements
    with pytest.raises(TypeError):
        zivid.Settings.RegionOfInterest.Depth(range=[])
    with pytest.raises(TypeError):
        zivid.Settings.RegionOfInterest.Depth(range=[100.0])
    with pytest.raises(TypeError):
        zivid.Settings.RegionOfInterest.Depth(range=[100.0, 200.0, 300.0])

    # A range must have min <= max
    with pytest.raises(TypeError):
        zivid.Settings.RegionOfInterest.Depth(range=[100.0, -200.0])


def test_print_settings(
    application,  # pylint: disable=unused-argument
):
    print(zivid.Settings())


def test_print_acquisition(
    application,  # pylint: disable=unused-argument
):
    print(zivid.Settings.Acquisition())


def test_print_processing(
    application,  # pylint: disable=unused-argument
):
    print(zivid.Settings.Processing())



"""Contains the UnorganizedPointCloud class."""

import _zivid
import numpy


class UnorganizedPointCloud:
    """Point cloud with x, y, z, RGB color and SNR laid out as a linear list of only valid points.

    An instance of this class is a handle to a point cloud stored on the compute device memory.
    This class provides several methods to copy point cloud data from the compute device
    memory to host (CPU) system memory (RAM).

    This point cloud contains only valid points, meaning that the XYZ values are never NaN.
    """

    def __init__(self, impl=None):
        """Create an empty point cloud.

        This constructor creates a point cloud with size zero, i.e. no points. An empty point cloud
        can be useful for combining points from several other points clouds.

        The argument impl is only used internally, and should not be set by the end-user.

        Args:
            impl:   Reference to internal/back-end instance.

        Raises:
            TypeError: If argument does not match the expected internal class.
        """
        if impl is None:
            self.__impl = _zivid.UnorganizedPointCloud()
        else:
            if not isinstance(impl, _zivid.UnorganizedPointCloud):
                raise TypeError(
                    "Unsupported type for argument impl. Got {}, expected {}".format(
                        type(impl), _zivid.UnorganizedPointCloud
                    )
                )
            self.__impl = impl

    @property
    def size(self):
        """Get the size of the point cloud (number of points)."""
        return self.__impl.size()

    def extended(self, other):
        """Create a new point cloud containing the combined data of this point cloud and another.

        Args:
            other: The other UnorganizedPointCloud to copy data from

        Returns:
            A new UnorganizedPointCloud containing the combined data
        """
        if not isinstance(other, UnorganizedPointCloud):
            raise TypeError(
                "Unsupported type for argument other. Got {}, expected {}".format(type(other), UnorganizedPointCloud)
            )
        return UnorganizedPointCloud(self.__impl.extended(other.__impl))  # pylint: disable=protected-access

    def extend(self, other):
        """Extend this point cloud in-place by adding the points from another point cloud.

        Args:
            other: The other UnorganizedPointCloud to copy data from

        Returns:
            Reference to the same UnorganizedPointCloud instance (for chaining calls)
        """
        if not isinstance(other, UnorganizedPointCloud):
            raise TypeError(
                "Unsupported type for argument other. Got {}, expected {}".format(type(other), UnorganizedPointCloud)
            )
        self.__impl.extend(other.__impl)  # pylint: disable=protected-access
        return self

    def voxel_downsampled(self, voxel_size, min_points_per_voxel):
        """Create a new point cloud that is a voxel downsampling of this point cloud.

        Voxel downsampling subdivides 3D space into a grid of cubic voxels with a given size. If a given voxel
        contains a number of points at or above the given limit, all those source points are replaced with a
        single new point at the SNR-weighted average position.

        Args:
            voxel_size: The size of the voxel cubes (must be greater than 0.0)
            min_points_per_voxel: The minimum number of points required to fill a voxel (must be 1 or greater)

        Returns:
            A new UnorganizedPointCloud containing the voxel downsampled data
        """
        return UnorganizedPointCloud(self.__impl.voxel_downsampled(voxel_size, min_points_per_voxel))

    def transform(self, matrix):
        """Transform the point cloud in-place by a 4x4 transformation matrix.

        The transform matrix must be affine, i.e., the last row of the matrix should be [0, 0, 0, 1].

        Args:
            matrix: A 4x4 numpy arrays of floats

        Returns:
            Reference to the same UnorganizedPointCloud instance (for chaining calls)
        """
        self.__impl.transform(matrix)
        return self

    def transformed(self, matrix):
        """Get a transformed copy of the point cloud.

        This method is identical to "transform", except the transformed point cloud is
        returned as a new UnorganizedPointCloud instance. The current point cloud is not modified.

        Args:
            matrix: A 4x4 numpy arrays of floats

        Returns:
            A new UnorganizedPointCloud instance
        """
        return UnorganizedPointCloud(self.__impl.transformed(matrix))

    def center(self):
        """Translate the point cloud in-place so that its centroid lands at the origin (0,0,0)."""
        self.__impl.center()
        return self

    def centroid(self):
        """Get the centroid of the point cloud, i.e. average of all XYZ point positions.

        Returns:
            The XYZ centroid as a numpy array, or None if the point cloud is empty.
        """
        return self.__impl.centroid()

    def copy_data(self, data_format):
        """Copy point cloud data from GPU to numpy array.

        Supported data formats:
        xyz:            ndarray(Size,3) of float
        rgba:           ndarray(Size,4) of uint8
        bgra:           ndarray(Size,4) of uint8
        rgba_srgb:      ndarray(Size,4) of uint8
        bgra_srgb:      ndarray(Size,4) of uint8
        snr:            ndarray(Size)   of float

        Args:
            data_format: A string specifying the data to be copied

        Returns:
            A numpy array with the requested data.

        Raises:
            ValueError: if the requested data format does not exist
        """
        self.__impl.assert_not_released()

        data_formats = {
            "xyz": _zivid.Array1DPointXYZ,
            "rgba": _zivid.Array1DColorRGBA,
            "bgra": _zivid.Array1DColorBGRA,
            "rgba_srgb": _zivid.Array1DColorRGBA_SRGB,
            "bgra_srgb": _zivid.Array1DColorBGRA_SRGB,
            "snr": _zivid.Array1DSNR,
        }
        try:
            data_format_class = data_formats[data_format]
        except KeyError as ex:
            raise ValueError(
                "Unsupported data format: {data_format}. Supported formats: {all_formats}".format(
                    data_format=data_format, all_formats=list(data_formats.keys())
                )
            ) from ex
        return numpy.array(data_format_class(self.__impl))

    def clone(self):
        """Get a clone of the point cloud.

        The clone will include a copy of all the point cloud data on the compute device memory. This means that the
        returned point cloud will not be affected by subsequent modifications (such as transform or downsample) on the
        original point cloud.

        This function incurs a performance cost due to the copying of the compute device memory. When performance is
        important we recommend to avoid using this method, and instead modify the existing point cloud.

        This method is equivalent to calling `copy.deepcopy` on the point cloud. You can obtain a shallow copy that does
        not copy the underlying data by using `copy.copy` on the point cloud instead.

        Returns:
            A new UnorganizedPointCloud instance
        """
        return UnorganizedPointCloud(self.__impl.clone())

    def release(self):
        """Release the underlying resources."""
        try:
            impl = self.__impl
        except AttributeError:
            pass
        else:
            impl.release()

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.release()

    def __del__(self):
        self.release()

    def __copy__(self):
        return UnorganizedPointCloud(self.__impl.__copy__())

    def __deepcopy__(self, memodict):
        return UnorganizedPointCloud(self.__impl.__deepcopy__(memodict))



"""Module for experimental point cloud registration. This API may change in the future."""

import _zivid
from zivid._local_point_cloud_registration_parameters import (
    LocalPointCloudRegistrationParameters,
    _to_internal_toolbox_local_point_cloud_registration_parameters,
)
from zivid.calibration import Pose
from zivid.matrix4x4 import Matrix4x4
from zivid.unorganized_point_cloud import UnorganizedPointCloud


class LocalPointCloudRegistrationResult:
    """The result of a call to local_point_cloud_registration."""

    def __init__(self, impl):
        """Initialize LocalPointCloudRegistrationResult wrapper.

        This constructor is only used internally, and should not be called by the end-user.

        Args:
            impl:   Reference to internal/back-end instance.

        Raises:
            TypeError: If argument does not match the expected internal class.
        """
        if not isinstance(impl, _zivid.toolbox.LocalPointCloudRegistrationResult):
            raise TypeError(
                "Unsupported type for argument impl. Got {}, expected {}".format(
                    type(impl), _zivid.toolbox.LocalPointCloudRegistrationResult
                )
            )
        self.__impl = impl

    def transform(self):
        """The transform that must be applied to the source point cloud for it to align with the target point cloud.

        Returns:
            A rigid transform as a Pose instance
        """
        return Pose(self.__impl.transform().to_matrix())

    def converged(self):
        """A boolean indicating whether the convergence criteria were satisfied before reaching the iteration limit.

        Returns:
            A boolean indicating convergence
        """
        return self.__impl.converged()

    def source_coverage(self):
        """The fraction of points in the source that has a correspondence in the target after transformation.

        Returns:
            A float in the range [0.0, 1.0]
        """
        return self.__impl.source_coverage()

    def root_mean_square_error(self):
        """The root mean squared distance between corresponding points in the source and target after transformation.

        Returns:
            A float representing the root mean square error
        """
        return self.__impl.root_mean_square_error()

    def __str__(self):
        return str(self.__impl)


def local_point_cloud_registration(target, source, parameters=None, initial_transform=None):
    """Compute alignment transform between two point clouds.

    Given a `source` point cloud and a `target` point cloud, this function attempts to compute the transform
    that must be applied to the `source` in order to align it with the `target`. This can be used to create a
    "stitched" unorganized point cloud of an object by combining data collected from different camera angles.

    This function takes an argument `initial_transform` which is used as a starting-point for the computation
    of the transform that best aligns `source` with `target`. This initial guess is usually found from e.g.
    reference markers or robot capture pose, and this function is then used to refine the alignment. If the
    overlap of `source` and `target` is already quite good, one can pass the identity pose as `initial_transform`.

    The returned transform represents the total transform needed to align `source` with `target`, i.e. it
    includes both `initial_transform` and the refinement found by the algorithm.

    Performance is very dependent on the number of points in either point cloud. To improve performance,
    voxel downsample one or both point clouds before passing them into this function. The resulting alignment
    transform can then be applied to the non-downsampled point clouds to still obtain a dense result.

    Performance is also very dependent on `MaxCorrespondenceDistance`. To improve performance, try reducing
    this value. However, keep the value larger than the typical point-to-point distance in the point clouds,
    and larger than the expected translation error in the initial guess.

    Args:
        target:               The target UnorganizedPointCloud to align with
        source:               The source UnorganizedPointCloud to be aligned with target
        parameters:           An optional LocalPointCloudRegistrationParameters instance.
                              If not provided, default parameters are used.
        initial_transform:    Optional transform (Pose/Matrix4x4/ndarray) be used as a starting point for the algorithm.
                              If not provided, the identity transform is used.

    Returns:
        A LocalPointCloudRegistrationResult instance
    """
    if not isinstance(target, UnorganizedPointCloud):
        raise TypeError(
            "Unsupported type for argument target. Got {}, expected {}".format(type(target), UnorganizedPointCloud)
        )
    if not isinstance(source, UnorganizedPointCloud):
        raise TypeError(
            "Unsupported type for argument source. Got {}, expected {}".format(type(source), UnorganizedPointCloud)
        )

    if parameters is None:
        parameters = LocalPointCloudRegistrationParameters()
    else:
        if not isinstance(parameters, LocalPointCloudRegistrationParameters):
            raise TypeError(
                "Unsupported type for argument parameters. Got {}, expected {}".format(
                    type(parameters), LocalPointCloudRegistrationParameters
                )
            )

    if initial_transform is None:
        initial_transform_pose = Pose(Matrix4x4.identity())
    elif isinstance(initial_transform, Pose):
        initial_transform_pose = initial_transform
    else:
        initial_transform_pose = Pose(initial_transform)

    return LocalPointCloudRegistrationResult(
        _zivid.toolbox.local_point_cloud_registration(
            target=target._UnorganizedPointCloud__impl,  # pylint: disable=protected-access
            source=source._UnorganizedPointCloud__impl,  # pylint: disable=protected-access
            params=_to_internal_toolbox_local_point_cloud_registration_parameters(parameters),
            initial_transform=initial_transform_pose._Pose__impl,  # pylint: disable=protected-access
        )
    )



import copy

import numpy as np
import pytest
import zivid
from assertions import assert_unorganized_point_clouds_equal, assert_unorganized_point_clouds_not_equal


def test_create_empty_unorganized_point_cloud(
    application,  # pylint: disable=unused-argument
):
    upc = zivid.UnorganizedPointCloud()
    assert isinstance(upc, zivid.UnorganizedPointCloud)
    assert upc.size == 0

    xyz = upc.copy_data("xyz")
    rgba = upc.copy_data("rgba")
    snr = upc.copy_data("snr")

    assert isinstance(xyz, np.ndarray)
    assert xyz.dtype == np.float32
    assert xyz.shape == (0, 3)

    assert isinstance(rgba, np.ndarray)
    assert rgba.dtype == np.uint8
    assert rgba.shape == (0, 4)

    assert isinstance(snr, np.ndarray)
    assert snr.dtype == np.float32
    assert snr.shape == (0,)


def test_create_unorganized_point_cloud(point_cloud):
    n_points_original = point_cloud.width * point_cloud.height
    xyz_organized = point_cloud.copy_data("xyz")
    rgba_organized = point_cloud.copy_data("rgba")
    bgra_organized = point_cloud.copy_data("bgra")
    rgba_srgb_organized = point_cloud.copy_data("rgba_srgb")
    bgra_srgb_organized = point_cloud.copy_data("bgra_srgb")
    snr_organized = point_cloud.copy_data("snr")

    valid_mask = ~np.isnan(xyz_organized).any(axis=2)
    n_points_valid = np.sum(valid_mask.flatten())
    assert n_points_valid < n_points_original

    xyz_unorganized_reference = xyz_organized[valid_mask]
    rgba_unorganized_reference = rgba_organized[valid_mask]
    bgra_unorganized_reference = bgra_organized[valid_mask]
    rgba_srgb_unorganized_reference = rgba_srgb_organized[valid_mask]
    bgra_srgb_unorganized_reference = bgra_srgb_organized[valid_mask]
    snr_unorganized_reference = snr_organized[valid_mask]

    upc = point_cloud.to_unorganized_point_cloud()
    assert isinstance(upc, zivid.UnorganizedPointCloud)
    assert upc.size == n_points_valid

    xyz = upc.copy_data("xyz")
    assert isinstance(xyz, np.ndarray)
    assert xyz.dtype == np.float32
    assert xyz.shape == (n_points_valid, 3)
    np.testing.assert_array_equal(xyz_unorganized_reference, xyz)

    rgba = upc.copy_data("rgba")
    assert isinstance(rgba, np.ndarray)
    assert rgba.dtype == np.uint8
    assert rgba.shape == (n_points_valid, 4)
    np.testing.assert_array_equal(rgba_unorganized_reference, rgba)

    bgra = upc.copy_data("bgra")
    assert isinstance(bgra, np.ndarray)
    assert bgra.dtype == np.uint8
    assert bgra.shape == (n_points_valid, 4)
    np.testing.assert_array_equal(bgra_unorganized_reference, bgra)

    rgba_srgb = upc.copy_data("rgba_srgb")
    assert isinstance(rgba_srgb, np.ndarray)
    assert rgba_srgb.dtype == np.uint8
    assert rgba_srgb.shape == (n_points_valid, 4)
    np.testing.assert_array_equal(rgba_srgb_unorganized_reference, rgba_srgb)

    bgra_srgb = upc.copy_data("bgra_srgb")
    assert isinstance(bgra_srgb, np.ndarray)
    assert bgra_srgb.dtype == np.uint8
    assert bgra_srgb.shape == (n_points_valid, 4)
    np.testing.assert_array_equal(bgra_srgb_unorganized_reference, bgra_srgb)

    snr = upc.copy_data("snr")
    assert isinstance(snr, np.ndarray)
    assert snr.dtype == np.float32
    assert snr.shape == (n_points_valid,)
    np.testing.assert_array_equal(snr_unorganized_reference, snr)


def test_unorganized_point_cloud_extended(point_cloud):
    upc1 = point_cloud.to_unorganized_point_cloud()
    upc2 = point_cloud.to_unorganized_point_cloud()

    upc_extended = upc1.extended(upc2)
    assert upc_extended.size == 2 * upc1.size

    xyz = upc1.copy_data("xyz")
    xyz_extended = upc_extended.copy_data("xyz")
    assert xyz_extended.shape == (2 * upc1.size, 3)
    np.testing.assert_array_equal(xyz_extended[: upc1.size], xyz)
    np.testing.assert_array_equal(xyz_extended[upc1.size :], xyz)


def test_unorganized_point_cloud_extend_empty(point_cloud):
    upc_canvas = zivid.UnorganizedPointCloud()
    assert upc_canvas.size == 0

    upc = point_cloud.to_unorganized_point_cloud()
    upc_size = upc.size
    assert upc_size > 0

    upc_canvas.extend(upc)
    assert upc.size == upc_size
    assert upc_canvas.size == upc_size

    upc_canvas.extend(upc)
    assert upc.size == upc_size
    assert upc_canvas.size == 2 * upc_size


def test_unorganized_point_cloud_voxel_downsampled(point_cloud):
    upc = point_cloud.to_unorganized_point_cloud()

    upc_downsampled = upc.voxel_downsampled(voxel_size=5.0, min_points_per_voxel=1)
    assert upc_downsampled.size < upc.size


def test_unorganized_point_cloud_transform(point_cloud):
    upc = point_cloud.to_unorganized_point_cloud()

    transform_matrix = np.array(
        [
            [1.0, 0.0, 0.0, 1.0],
            [0.0, 1.0, 0.0, 2.0],
            [0.0, 0.0, 1.0, 3.0],
            [0.0, 0.0, 0.0, 1.0],
        ]
    )

    xyz_before = upc.copy_data("xyz")
    upc.transform(transform_matrix)
    xyz_after = upc.copy_data("xyz")
    np.testing.assert_allclose(xyz_after[:, :3], xyz_before[:, :3] + np.array([1.0, 2.0, 3.0]))


def test_unorganized_point_cloud_transformed(point_cloud):
    upc = point_cloud.to_unorganized_point_cloud()

    transform_matrix = np.array(
        [
            [1.0, 0.0, 0.0, 1.0],
            [0.0, 1.0, 0.0, 2.0],
            [0.0, 0.0, 1.0, 3.0],
            [0.0, 0.0, 0.0, 1.0],
        ]
    )

    # Get original points
    xyz_before = upc.copy_data("xyz")

    # Get transformed copy
    upc_transformed = upc.transformed(transform_matrix)
    assert isinstance(upc_transformed, zivid.UnorganizedPointCloud)
    assert upc_transformed is not upc

    # Original point cloud should not be modified
    np.testing.assert_array_equal(xyz_before, upc.copy_data("xyz"))

    # New point cloud should have the transformation applied
    xyz_after = upc_transformed.copy_data("xyz")
    np.testing.assert_allclose(xyz_after[:, :3], xyz_before[:, :3] + np.array([1.0, 2.0, 3.0]))


def test_unorganized_point_cloud_center_and_centroid(point_cloud):
    upc = point_cloud.to_unorganized_point_cloud().voxel_downsampled(0.5, 1)
    assert upc.size > 1_000_000, "Must test on large point clouds to check for float precision bugs"

    centroid_before = upc.centroid()
    assert isinstance(centroid_before, np.ndarray)
    xyz_before = upc.copy_data("xyz")
    centroid_before_reference = np.mean(xyz_before, axis=0, dtype=np.float64)

    np.testing.assert_allclose(centroid_before, centroid_before_reference, rtol=1e-4)

    upc.center()
    centroid_after = upc.centroid()
    xyz_after = upc.copy_data("xyz")
    centroid_after_reference = np.mean(xyz_after, axis=0, dtype=np.float64)

    np.testing.assert_allclose(centroid_after, [0.0, 0.0, 0.0], atol=1e-3)
    np.testing.assert_allclose(centroid_after_reference, [0.0, 0.0, 0.0], atol=1e-3)


def test_unorganized_point_cloud_center_and_centroid_empty(point_cloud):
    upc_empty = point_cloud.to_unorganized_point_cloud().voxel_downsampled(1.0, 10000)
    assert upc_empty.size == 0

    centroid = upc_empty.centroid()
    assert centroid is None

    with pytest.raises(RuntimeError):
        upc_empty.center()


def test_illegal_init(
    application,  # pylint: disable=unused-argument
):
    with pytest.raises(TypeError):
        zivid.UnorganizedPointCloud("Should fail.")

    with pytest.raises(TypeError):
        zivid.UnorganizedPointCloud(123)


def test_copy_unorganized_point_cloud(point_cloud, transform):
    upc = point_cloud.to_unorganized_point_cloud()

    with copy.copy(upc) as upc_copy:
        assert isinstance(upc_copy, zivid.UnorganizedPointCloud)
        assert upc_copy is not upc
        assert_unorganized_point_clouds_equal(upc, upc_copy)

        upc.transform(transform)
        # shallow copy, transform should have affected both copies
        assert_unorganized_point_clouds_equal(upc, upc_copy)


def test_deepcopy_unorganized_point_cloud(point_cloud, transform):
    upc = point_cloud.to_unorganized_point_cloud()

    with copy.deepcopy(upc) as upc_copy:
        assert isinstance(upc_copy, zivid.UnorganizedPointCloud)
        assert upc_copy is not upc
        assert_unorganized_point_clouds_equal(upc, upc_copy)

        upc.transform(transform)
        # deep copy, transform should not have affected the copy
        assert_unorganized_point_clouds_not_equal(upc, upc_copy)


def test_clone_point_cloud(point_cloud, transform):
    upc = point_cloud.to_unorganized_point_cloud()

    with upc.clone() as upc_clone:
        assert isinstance(upc_clone, zivid.UnorganizedPointCloud)
        assert upc_clone is not upc
        assert_unorganized_point_clouds_equal(upc, upc_clone)

        upc.transform(transform)
        # clone, transform should not have affected the clone
        assert_unorganized_point_clouds_not_equal(upc, upc_clone)



import copy
import tempfile
from pathlib import Path

import numpy as np
import pytest
import zivid


def test_copy_data(image_2d):
    data = image_2d.copy_data()
    assert data is not None
    assert isinstance(data, np.ndarray)
    assert data.shape == (image_2d.height, image_2d.width, 4)
    assert data.dtype == np.uint8


def test_save_path(image_2d):
    with tempfile.TemporaryDirectory() as tmpdir:
        image_file = Path(tmpdir) / "some_file.png"
        image_2d.save(image_file)


def test_save_string(image_2d):
    with tempfile.TemporaryDirectory() as tmpdir:
        image_file = str(Path(tmpdir) / "some_file.png")
        image_2d.save(image_file)


def test_to_array_context_manager(frame_2d, color_format):
    with getattr(frame_2d, f"image_{color_format}")() as image_2d:
        image_2d.copy_data()
    with pytest.raises(RuntimeError):
        image_2d.copy_data()


def test_save_context_manager(frame_2d, color_format):
    with tempfile.TemporaryDirectory() as tmpdir:
        image_file = str(Path(tmpdir) / "some_file.png")
        with getattr(frame_2d, f"image_{color_format}")() as image_2d:
            image_2d.save(image_file)
        with pytest.raises(RuntimeError):
            image_2d.save(image_file)


def test_height(image_2d):
    height = image_2d.height

    assert height is not None
    assert isinstance(height, int)


def test_width(image_2d):
    width = image_2d.width

    assert width is not None
    assert isinstance(width, int)


def test_load(image_2d, color_format):
    with tempfile.TemporaryDirectory() as tmpdir:
        image_file = Path(tmpdir) / "saved_image.png"
        image_2d.save(image_file)

        loaded_image = zivid.Image.load(image_file, color_format=color_format)
        np.testing.assert_array_equal(image_2d.copy_data(), loaded_image.copy_data())


def test_load_invalid_color_format(frame_2d):
    with tempfile.TemporaryDirectory() as tmpdir:
        image_file = Path(tmpdir) / "saved_image.png"
        frame_2d.image_rgba().save(image_file)
        with pytest.raises(ValueError):
            zivid.Image.load(image_file, "bgr")
        with pytest.raises(ValueError):
            zivid.Image.load(image_file, "rgb")
        with pytest.raises(ValueError):
            zivid.Image.load(image_file, "asdf")


def test_copy(frame_2d, color_format):
    with getattr(frame_2d, f"image_{color_format}")() as image_2d:
        with copy.copy(image_2d) as copied_image:
            assert copied_image is not None
            assert isinstance(copied_image, zivid.Image)
            assert copied_image is not image_2d
            np.testing.assert_array_equal(image_2d.copy_data(), copied_image.copy_data())



import _zivid
from zivid.experimental.point_cloud_export.file_format import PCD, PLY, XYZ
from zivid.unorganized_point_cloud import UnorganizedPointCloud


def export_unorganized_point_cloud(unorganized_point_cloud, file_format):
    """Save UnorganizedPointCloud to a file.

    The file format is specified by the file_format argument. The file format can be PLY, XYZ, or
    PCD.

    The layout for PLY file format must be `unordered` and layout for PCD file format must be
    `unorganized`.

    Args:
        unorganized_point_cloud: Unorganized point cloud to export.
        file_format: File format specification.

    Raises:
        TypeError: If unorganized_point_cloud is not a UnorganizedPointCloud.
        TypeError: If file_format is not a file format specification.
    """
    if not isinstance(unorganized_point_cloud, UnorganizedPointCloud):
        raise TypeError(
            "Unsupported type for argument frame. Got {}, expected {}".format(
                type(unorganized_point_cloud), UnorganizedPointCloud
            )
        )
    if not any(
        [
            isinstance(file_format, PLY),
            isinstance(file_format, XYZ),
            isinstance(file_format, PCD),
        ]
    ):
        raise TypeError(
            "Unsupported type for argument file_format. Got {}, expected {}".format(
                type(file_format),
                " or ".join([t.__name__ for t in [PLY, XYZ, PCD]]),
            )
        )

    format_impl_attr = f"_{type(file_format).__name__}__impl"
    _zivid.point_cloud_export.export_unorganized_point_cloud(
        unorganized_point_cloud._UnorganizedPointCloud__impl,  # pylint: disable=protected-access
        getattr(file_format, format_impl_attr),
    )



import tempfile

import pytest


def test_capture_from_file():
    with tempfile.TemporaryDirectory() as temp_dir:
        pytest.helpers.run_sample(name="capture_from_file", working_directory=temp_dir)


def test_print_version_info():
    pytest.helpers.run_sample(name="print_version_info")


@pytest.mark.physical_camera
def test_capture():
    with tempfile.TemporaryDirectory() as temp_dir:
        pytest.helpers.run_sample(name="capture", working_directory=temp_dir)


@pytest.mark.physical_camera
def test_capture_2d():
    with tempfile.TemporaryDirectory() as temp_dir:
        pytest.helpers.run_sample(name="capture_2d", working_directory=temp_dir)


@pytest.mark.physical_camera
def test_capture_assistant():
    with tempfile.TemporaryDirectory() as temp_dir:
        pytest.helpers.run_sample(name="capture_assistant", working_directory=temp_dir)


@pytest.mark.physical_camera
def test_capture_hdr():
    with tempfile.TemporaryDirectory() as temp_dir:
        pytest.helpers.run_sample(name="capture_hdr", working_directory=temp_dir)



"""Module for exporting point cloud data to various formats. This API may change in the future."""

from zivid.experimental.point_cloud_export._export_frame import export_frame
from zivid.experimental.point_cloud_export._export_unorganized_point_cloud import export_unorganized_point_cloud



import pytest
import zivid

# These tests cannot be in the "test_application" module because that module depends on a
# zivid.Application fixture and these tests want to test the initialization and teardown of the
# zivid.Application class.


def test_init_with():
    with zivid.Application() as app:
        assert app
        assert isinstance(app, zivid.Application)


def test_release(file_camera_file):
    application = zivid.Application()
    assert application.create_file_camera(file_camera_file)
    application.release()
    with pytest.raises(RuntimeError):
        application.create_file_camera(file_camera_file)


def test_delete(file_camera_file):
    app1 = zivid.Application()
    assert app1.create_file_camera(file_camera_file)
    del app1

    app2 = zivid.Application()
    assert app2.create_file_camera(file_camera_file)



"""Module containing implementation of feature point detection functionality.

This module should not be imported directly by end-user, but rather accessed through
the zivid.calibration module.
"""

import _zivid
import numpy
from zivid._calibration.pose import Pose
from zivid.camera import Camera
from zivid.frame import Frame


class CalibrationBoardDetectionStatus:  # pylint: disable=too-few-public-methods
    """Enumeration of possible detection statuses.

    The DetectionResult.status() method will return one of these values.
    """

    ok = "ok"
    no_valid_fiducial_marker_detected = "no_valid_fiducial_marker_detected"
    multiple_valid_fiducial_markers_detected = "multiple_valid_fiducial_markers_detected"
    board_detection_failed = "board_detection_failed"
    insufficient_3d_quality = "insufficient_3d_quality"


class DetectionResult:
    """Class representing detected feature points from a calibration board."""

    def __init__(self, impl):
        """Initialize DetectionResult wrapper.

        This constructor is only used internally, and should not be called by the end-user.

        Args:
            impl:   Reference to internal/back-end instance.

        Raises:
            TypeError: If argument does not match the expected internal class.
        """
        if not isinstance(impl, _zivid.calibration.DetectionResult):
            raise TypeError(
                "Unsupported type for argument impl. Got {}, expected {}".format(
                    type(impl), _zivid.calibration.DetectionResult
                )
            )

        self.__impl = impl

    def valid(self):
        """Check validity of DetectionResult.

        Returns:
            True if DetectionResult is valid
        """
        return self.__impl.valid()

    def centroid(self):
        """Get the centroid of the detected feature points.

        Will throw an exception if the DetectionResult is not valid.

        Returns:
            A 1D array containing the X, Y and Z coordinates of the centroid
        """
        return self.__impl.centroid()

    def pose(self):
        """Get position and orientation of the top left detected corner in camera-space.

        Pose calculation works for official Zivid calibration boards only.
        An exception will be thrown if valid() is false or if the board is not supported.

        Returns:
            The Pose of the top left corner (4x4 transformation matrix)
        """
        return Pose(self.__impl.pose().to_matrix())

    def status(self):
        """Get the status of the detection.

        Returns:
            The detection status as a string. See CalibrationBoardDetectionStatus for possible values.
        """
        return self.__impl.status().name

    def status_description(self):
        """Get a human-readable description of the status.

        Useful for feedback if valid() returns False.

        This returns a free-form string and should not be considered API stable.
        Use status() for programmatic checks.

        Returns:
            A string which describes the detection status.
        """
        return self.__impl.status_description()

    def feature_points(self):
        """Get the detected feature points in camera-space.

        Returns a 2D array of 3D coordinates representing the centers of the calibration
        board squares. The element at index (0,0) corresponds to the physical top-left
        corner of the board. The first index moves down the physical board, while the second
        index moves to the right. If the valid() function returns false, an exception will
        be thrown.

        Returns:
            Feature points.
        """
        return self.__impl.feature_points()

    def feature_points_2d(self):
        """Get the detected feature points in pixel-space.

        Same as feature_points(), but with 2D coordinates instead of 3D coordinates. The points
        are reported at subpixel accuracy. If the valid() function returns false, an exception
        will be thrown.

        Returns:
            Feature points in 2D.
        """
        return self.__impl.feature_points_2d()

    def __bool__(self):
        return bool(self.__impl)

    def __str__(self):
        return str(self.__impl)


class MarkerShape:
    """Holds physical (3D) and image (2D) properties of a detected fiducial marker."""

    def __init__(self, impl):
        """Initialize MarkerShape wrapper.

        This constructor is only used internally, and should not be called by the end-user.

        Args:
            impl:   Reference to internal/back-end instance.

        Raises:
            TypeError: If argument does not match the expected internal class.
        """
        if not isinstance(impl, _zivid.calibration.MarkerShape):
            raise TypeError(
                "Unsupported type for argument impl. Got {}, expected {}".format(
                    type(impl), _zivid.calibration.MarkerShape
                )
            )

        self.__impl = impl

    @property
    def corners_in_pixel_coordinates(self):
        """Get 2D image coordinates of the corners of the detected marker.

        Returns:
            Four 2D corner coordinates as a 4x2 numpy array
        """
        return numpy.array(self.__impl.corners_in_pixel_coordinates())

    @property
    def corners_in_camera_coordinates(self):
        """Get 3D spatial coordinates of the corners of the detected marker.

        Returns:
            Four 3D corner coordinates as a 4x3 numpy array
        """
        return numpy.array(self.__impl.corners_in_camera_coordinates())

    @property
    def identifier(self):
        """Get the id of the detected marker.

        Returns:
            Id as int
        """
        return self.__impl.id_()

    @property
    def pose(self):
        """Get 3D pose of the marker.

        The returned pose will be positioned at the center of the marker, and have an orientation such that its z-axis
        points perpendicularly into the face of the marker.

        Returns:
            The Pose of the marker center (4x4 transformation matrix)
        """
        return Pose(self.__impl.pose().to_matrix())


class MarkerDictionary:
    """Holds information about fiducial markers such as ArUco or AprilTag for detection.

    This class's properties describe the different dictionaries available, for example
    aruco4x4_50 describes the ArUco dictionary with 50 markers of size 4x4.

    For more information on ArUco markers see the OpenCV documentation on ArUco markers:
    https://docs.opencv.org/4.x/d5/dae/tutorial_aruco_detection.html,

    To get more information about fiducial markers in general, refer to the wikipedia page:
    https://en.wikipedia.org/wiki/Fiducial_marker
    """

    aruco4x4_50 = "aruco4x4_50"
    aruco4x4_100 = "aruco4x4_100"
    aruco4x4_250 = "aruco4x4_250"
    aruco4x4_1000 = "aruco4x4_1000"
    aruco5x5_50 = "aruco5x5_50"
    aruco5x5_100 = "aruco5x5_100"
    aruco5x5_250 = "aruco5x5_250"
    aruco5x5_1000 = "aruco5x5_1000"
    aruco6x6_50 = "aruco6x6_50"
    aruco6x6_100 = "aruco6x6_100"
    aruco6x6_250 = "aruco6x6_250"
    aruco6x6_1000 = "aruco6x6_1000"
    aruco7x7_50 = "aruco7x7_50"
    aruco7x7_100 = "aruco7x7_100"
    aruco7x7_250 = "aruco7x7_250"
    aruco7x7_1000 = "aruco7x7_1000"

    _valid_values = {
        "aruco4x4_50": _zivid.calibration.MarkerDictionary.aruco4x4_50,
        "aruco4x4_100": _zivid.calibration.MarkerDictionary.aruco4x4_100,
        "aruco4x4_250": _zivid.calibration.MarkerDictionary.aruco4x4_250,
        "aruco4x4_1000": _zivid.calibration.MarkerDictionary.aruco4x4_1000,
        "aruco5x5_50": _zivid.calibration.MarkerDictionary.aruco5x5_50,
        "aruco5x5_100": _zivid.calibration.MarkerDictionary.aruco5x5_100,
        "aruco5x5_250": _zivid.calibration.MarkerDictionary.aruco5x5_250,
        "aruco5x5_1000": _zivid.calibration.MarkerDictionary.aruco5x5_1000,
        "aruco6x6_50": _zivid.calibration.MarkerDictionary.aruco6x6_50,
        "aruco6x6_100": _zivid.calibration.MarkerDictionary.aruco6x6_100,
        "aruco6x6_250": _zivid.calibration.MarkerDictionary.aruco6x6_250,
        "aruco6x6_1000": _zivid.calibration.MarkerDictionary.aruco6x6_1000,
        "aruco7x7_50": _zivid.calibration.MarkerDictionary.aruco7x7_50,
        "aruco7x7_100": _zivid.calibration.MarkerDictionary.aruco7x7_100,
        "aruco7x7_250": _zivid.calibration.MarkerDictionary.aruco7x7_250,
        "aruco7x7_1000": _zivid.calibration.MarkerDictionary.aruco7x7_1000,
    }

    @classmethod
    def valid_values(cls):
        """Get valid values for MarkerDictionary.

        Returns:
            A list of strings representing valid values for MarkerDictionary.
        """
        return list(cls._valid_values.keys())

    @classmethod
    def marker_count(cls, dictionary_name):
        """Get the number of markers in a dictionary.

        Args:
            dictionary_name: Name of the dictionary, e.g. "aruco4x4_50". Must be one of the values returned by
                valid_values().

        Returns:
            Number of markers in the dictionary.

        Raises:
            ValueError: If the dictionary name is not one of the valid values returned by
                valid_values().
        """
        if dictionary_name not in cls._valid_values:
            raise ValueError(
                "Invalid dictionary name '{}'. Valid values are {}".format(dictionary_name, cls.valid_values())
            )

        return cls._valid_values[dictionary_name].marker_count()


class DetectionResultFiducialMarkers:
    """Class representing detected fiducial markers."""

    def __init__(self, impl):
        """Initialize DetectionResultFiducialMarkers wrapper.

        This constructor is only used internally, and should not be called by the end-user.

        Args:
            impl:   Reference to internal/back-end instance.

        Raises:
            TypeError: If argument does not match the expected internal class.
        """
        if not isinstance(impl, _zivid.calibration.DetectionResultFiducialMarkers):
            raise TypeError(
                "Unsupported type for argument impl. Got {}, expected {}".format(
                    type(impl), _zivid.calibration.DetectionResultFiducialMarkers
                )
            )

        self.__impl = impl

    def valid(self):
        """Check validity of DetectionResult.

        Returns:
            True if DetectionResult is valid
        """
        return self.__impl.valid()

    def allowed_marker_ids(self):
        """Get the allowed marker ids this detection result was made with.

        Returns:
            A list of integers, equal to what was passed to the detection function.
        """
        return self.__impl.allowed_marker_ids()

    def detected_markers(self):
        """Get all detected markers.

        Returns:
            A list of MarkerShape instances
        """
        return [MarkerShape(impl) for impl in self.__impl.detected_markers()]

    def __bool__(self):
        return bool(self.__impl)

    def __str__(self):
        return str(self.__impl)


def detect_feature_points(point_cloud):
    """Detect feature points from a calibration object in a point cloud.

    Args:
        point_cloud: PointCloud containing a calibration object

    Returns:
        A DetectionResult instance
    """
    return DetectionResult(
        _zivid.calibration.detect_feature_points(point_cloud._PointCloud__impl)  # pylint: disable=protected-access
    )


def detect_calibration_board(source):
    """Detect feature points from a calibration board in a frame or using a given camera.

    If a camera is used, this function will perform a relatively slow but high-quality point cloud
    capture with the camera. This function is necessary for applications that require very
    high-accuracy DetectionResult, such as in-field verification/correction.

    The functionality is to be exclusively used in combination with Zivid verified calibration boards.
    For further information please visit https://support.zivid.com.

    Args:
        source: A frame containing an image of a calibration board or a camera pointed at
            a calibration board

    Raises:
        TypeError: If source is not of type Camera or Frame

    Returns:
        A DetectionResult instance
    """
    if isinstance(source, Camera):
        return DetectionResult(
            _zivid.calibration.detect_calibration_board(source._Camera__impl)  # pylint: disable=protected-access
        )
    if isinstance(source, Frame):
        return DetectionResult(
            _zivid.calibration.detect_calibration_board(source._Frame__impl)  # pylint: disable=protected-access
        )
    raise TypeError(
        "Unsupported type for argument source. Got {}, expected one of {}".format(
            type(source),
            (Camera, Frame),
        )
    )


def capture_calibration_board(camera):
    """Capture a calibration board with the given camera.

    The functionality is to be exclusively used in combination with Zivid verified calibration boards.
    For further information please visit https://support.zivid.com.

    This function will perform a relatively slow but high-quality point cloud capture with the
    given camera. This function is necessary for applications that require very high-accuracy
    captures, such as in-field verification/correction.

    The Frame that is returned from this function may be used as input to `detect_calibration_board`.
    You may also use `detect_calibration_board` directly, which will invoke this function under the
    hood and yield a DetectionResult.

    Args:
        camera: a Camera pointed at a calibration board

    Returns:
        A Frame
    """
    return Frame(_zivid.calibration.capture_calibration_board(camera._Camera__impl))  # pylint: disable=protected-access


def detect_markers(frame, allowed_marker_ids, marker_dictionary):
    """Detect fiducial markers such as ArUco markers in a frame.

    Only markers with integer IDs are supported. To get more information about fiducial markers, refer to the
    wikipedia page:  https://en.wikipedia.org/wiki/Fiducial_marker

    For more information on ArUco markers specifically, see the OpenCV documentation on ArUco markers:
    https://docs.opencv.org/4.x/d5/dae/tutorial_aruco_detection.html,

    Frame need not contain all markers listed in allowedMarkerIds for a successful detection.

    Args:
        frame: A frame containing an image of one or several fiducial markers
        allowed_marker_ids: List of the IDs of markers to be detected
        marker_dictionary: The name of the marker dictionary to use. The name must be one of the values returned by
            MarkerDictionary.valid_values()

    Raises:
        ValueError: If marker_dictionary is not one of the valid values returned by MarkerDictionary.valid_values()

    Returns:
        A DetectionResultFiducialMarkers instance
    """
    if marker_dictionary not in MarkerDictionary.valid_values():
        raise ValueError(
            "Invalid marker dictionary '{}'. Valid values are {}".format(
                marker_dictionary, MarkerDictionary.valid_values()
            )
        )
    dictionary = MarkerDictionary._valid_values.get(marker_dictionary)  # pylint: disable=protected-access

    return DetectionResultFiducialMarkers(
        _zivid.calibration.detect_markers(
            frame._Frame__impl,  # pylint: disable=protected-access
            allowed_marker_ids,
            dictionary,
        )
    )
